@@ -9,7 +9,6 @@ needed to control the in-game agent through ai2thor.server.
 import atexit
 from collections import deque, defaultdict
 from itertools import product
-import io
 import json
 import copy
 import logging
@@ -28,19 +27,18 @@ import uuid
 import tty
 import sys
 import termios
-import fcntl
 
-import zipfile
 
 import numpy as np
 import ai2thor.docker
-import ai2thor.downloader
 import ai2thor.wsgi_server
 import ai2thor.fifo_server
 from ai2thor.interact import InteractiveControllerPrompt, DefaultActions
 from ai2thor.server import DepthFormat
 from ai2thor.build import COMMIT_ID
+import ai2thor.build
 from ai2thor._quality_settings import QUALITY_SETTINGS, DEFAULT_QUALITY
+from ai2thor.util import makedirs
 
 import warnings
 
@@ -348,11 +346,6 @@ def process_alive(pid):
 
     return True
 
-# python2.7 compatible makedirs
-def makedirs(directory):
-    if not os.path.isdir(directory):
-        os.makedirs(directory)
-
 def distance(point1, point2):
     x_diff = (point1['x'] - point2['x']) ** 2
     z_diff = (point1['z'] - point2['z']) ** 2
@@ -372,6 +365,7 @@ class Controller(object):
             port=0,
             start_unity=True,
             local_executable_path=None,
+            local_build=False,
             width=300,
             height=300,
             x_display=None,
@@ -384,7 +378,7 @@ class Controller(object):
             add_depth_noise=False,
             download_only=False,
             include_private_scenes=False,
-            server_class=ai2thor.wsgi_server.WsgiServer,
+            server_class=ai2thor.fifo_server.FifoServer,
             **unity_initialization_parameters
     ):
         self.receptacle_nearest_pivot_points = {}
@@ -392,7 +386,9 @@ class Controller(object):
         self.unity_pid = None
         self.docker_enabled = docker_enabled
         self.container_id = None
-        self.local_executable_path = local_executable_path
+        self.width = width
+        self.height = height
+
         self.last_event = None
         self._scenes_in_build = None
         self.killing_unity = False
@@ -404,7 +400,7 @@ class Controller(object):
         self.add_depth_noise = add_depth_noise
         self.include_private_scenes = include_private_scenes
         self.server_class = server_class
-
+        self._build = None
 
         self.interactive_controller = InteractiveControllerPrompt(
             list(DefaultActions),
@@ -413,8 +409,16 @@ class Controller(object):
             image_per_frame=save_image_per_frame
         )
 
+        if local_executable_path:
+            self._build = ai2thor.build.ExternalBuild(local_executable_path)
+        else:
+            self._build = self.find_build(local_build)
+
+        if self._build is None:
+            raise Exception("Couldn't find a suitable build for platform: %s" % platform.system())
+
         if download_only:
-            self.download_binary()
+            self._build.download()
         else:
 
             self.start(
@@ -436,6 +440,9 @@ class Controller(object):
                     DeprecationWarning
                 )
 
+            if 'fastActionEmit' in self.initialization_parameters and self.server_class != ai2thor.fifo_server.FifoServer:
+                warnings.warn("fastAtionEmit is only available with the FifoServer");
+
             if 'continuousMode' in self.initialization_parameters:
                 warnings.warn(
                     "Warning: 'continuousMode' is deprecated and will be ignored,"
@@ -449,7 +456,7 @@ class Controller(object):
                 init_return = event.metadata['actionReturn']
                 self.server.set_init_params(init_return)
 
-                print("Initialize return: {}".format(init_return))
+                logging.info("Initialize return: {}".format(init_return))
             else:
                 raise RuntimeError('Initialize action failure: {}'.format(
                     event.metadata['errorMessage'])
@@ -495,7 +502,7 @@ class Controller(object):
 
         return self._scenes_in_build
 
-    def reset(self, scene='FloorPlan_Train1_1'):
+    def reset(self, scene='FloorPlan_Train1_1', **init_params):
         if re.match(r'^FloorPlan[0-9]+$', scene):
             scene = scene + "_physics"
 
@@ -514,7 +521,24 @@ class Controller(object):
 
         self.server.send(dict(action='Reset', sceneName=scene, sequenceId=0))
         self.last_event = self.server.receive()
-        
+
+        # update the initialization parameters
+        init_params = init_params.copy()
+
+        # width and height are updates in 'ChangeResolution', not 'Initialize'
+        if ('width' in init_params and  init_params['width'] != self.width) or (
+            'height' in init_params and init_params['height'] != self.height
+        ):
+            if 'width' in init_params:
+                self.width = init_params['width']
+                del init_params['width']
+            if 'height' in init_params:
+                self.height = init_params['height']
+                del init_params['height']
+            self.step(action='ChangeResolution', x=self.width, y=self.height)
+
+        # updates the initialization parameters
+        self.initialization_parameters.update(init_params)
         self.last_event = self.step(action='Initialize', **self.initialization_parameters)
 
         return self.last_event
@@ -528,29 +552,7 @@ class Controller(object):
             max_num_repeats=1,
             remove_prob=0.5):
 
-        if random_seed is None:
-            random_seed = random.randint(0, 2**32)
-
-        exclude_object_ids = []
-
-        for obj in self.last_event.metadata['objects']:
-            pivot_points = self.receptacle_nearest_pivot_points
-            # don't put things in pot or pan currently
-            if (pivot_points and obj['receptacle'] and
-                    pivot_points[obj['objectId']].keys()) or obj['objectType'] in ['Pot', 'Pan']:
-
-                #print("no visible pivots for receptacle %s" % o['objectId'])
-                exclude_object_ids.append(obj['objectId'])
-
-        return self.step(dict(
-            action='RandomInitialize',
-            randomizeOpen=randomize_open,
-            uniquePickupableObjectTypes=unique_object_types,
-            excludeObjectIds=exclude_object_ids,
-            excludeReceptacleObjectPairs=exclude_receptacle_object_pairs,
-            maxNumRepeats=max_num_repeats,
-            removeProb=remove_prob,
-            randomSeed=random_seed))
+        raise Exception("RandomInitialize has been removed.  Use InitialRandomSpawn - https://ai2thor.allenai.org/ithor/documentation/actions/initialization/#object-position-randomization")
 
     def scene_names(self):
         scenes = []
@@ -560,31 +562,34 @@ class Controller(object):
 
         return scenes
 
-    def unlock_release(self):
-        if self.lock_file:
-            fcntl.flock(self.lock_file, fcntl.LOCK_UN)
-
-    def lock_release(self):
-        build_dir = os.path.join(self.releases_dir(), self.build_name())
-        if os.path.isdir(build_dir):
-            self.lock_file = open(os.path.join(build_dir, ".lock"), "w")
-            fcntl.flock(self.lock_file, fcntl.LOCK_SH)
-
     def prune_releases(self):
-        current_exec_path = self.executable_path()
-        for d in os.listdir(self.releases_dir()):
-            release = os.path.join(self.releases_dir(), d)
+        current_exec_path = self._build.executable_path
+        rdir = self.releases_dir
+        makedirs(self.tmp_dir)
+        makedirs(self.releases_dir)
 
+        # sort my mtime ascending, keeping the 3 most recent, attempt to prune anything older
+        all_dirs = list(filter(os.path.isdir, map(lambda x: os.path.join(rdir, x), os.listdir(rdir))))
+        sorted_dirs = sorted(all_dirs, key=lambda x: os.stat(x).st_mtime)[:-3]
+        for release in sorted_dirs:
             if current_exec_path.startswith(release):
                 continue
+            try:
+                lf_path = release + ".lock"
+                lf = os.open(lf_path, os.O_RDWR | os.O_CREAT)
+                # we must try to get a lock here since its possible that a process could still
+                # be running with this release
+                fcntl.lockf(lf, fcntl.LOCK_EX | fcntl.LOCK_NB)
 
-            if os.path.isdir(release):
-                try:
-                    with open(os.path.join(release, ".lock"), "w") as f:
-                        fcntl.flock(f, fcntl.LOCK_EX | fcntl.LOCK_NB)
-                        shutil.rmtree(release)
-                except Exception:
-                    pass
+                tmp_prune_dir = os.path.join(self.tmp_dir, '-'.join([self.build_name(release), str(time.time()), str(random.random()), 'prune']))
+                os.rename(release, tmp_prune_dir)
+                shutil.rmtree(tmp_prune_dir)
+
+                fcntl.lockf(lf, fcntl.LOCK_UN)
+                os.close(lf)
+                os.unlink(lf_path)
+            except Exception as e:
+                pass
 
     def next_interact_command(self):
         current_buffer = ''
@@ -665,10 +670,12 @@ class Controller(object):
         if ('objectId' in action and (action['action'] == 'OpenObject' or action['action'] == 'CloseObject')):
 
             force_visible = action.get('forceVisible', False)
-            if not force_visible and self.last_event.instance_detections2D and action['objectId'] not in self.last_event.instance_detections2D:
+            agent_id = action.get('agentId', 0)
+            instance_detections2D = self.last_event.events[agent_id].instance_detections2D
+            if not force_visible and instance_detections2D and action['objectId'] not in instance_detections2D:
                 should_fail = True
 
-            obj_metadata = self.last_event.get_object(action['objectId'])
+            obj_metadata = self.last_event.events[agent_id].get_object(action['objectId'])
             if obj_metadata is None or obj_metadata['isOpen'] == (action['action'] == 'OpenObject'):
                 should_fail = True
 
@@ -688,7 +695,7 @@ class Controller(object):
         self.server.send(action)
         self.last_event = self.server.receive()
 
-        if not self.last_event.metadata['lastActionSuccess'] and self.last_event.metadata['errorCode'] in ['InvalidAction', 'MissingArguments']:
+        if not self.last_event.metadata['lastActionSuccess'] and self.last_event.metadata['errorCode'] in ['InvalidAction', 'MissingArguments', 'AmbiguousAction']:
             raise ValueError(self.last_event.metadata['errorMessage'])
 
         if raise_for_failure:
@@ -697,7 +704,7 @@ class Controller(object):
         return self.last_event
 
     def unity_command(self, width, height, headless):
-        command = self.executable_path()
+        command = self._build.executable_path
         if headless:
             command += " -batchmode"
         else:
@@ -745,110 +752,50 @@ class Controller(object):
                 assert subprocess.call("xdpyinfo", stdout=dn, env=env, shell=True) == 0, \
                     ("Invalid DISPLAY %s - cannot find X server with xdpyinfo" % x_display)
 
-    def releases_dir(self):
-        return os.path.join(self.base_dir(), 'releases')
+    @property
+    def tmp_dir(self):
+        return os.path.join(self.base_dir, 'tmp')
 
+    @property
+    def releases_dir(self):
+        return os.path.join(self.base_dir, 'releases')
+
+    @property
     def base_dir(self):
         return os.path.join(os.path.expanduser('~'), '.ai2thor')
 
-    def build_url(self):
+    def find_build(self, local_build):
         from ai2thor.build import arch_platform_map
         import ai2thor.build
         arch = arch_platform_map[platform.system()]
+
         if COMMIT_ID:
-            ver_build = ai2thor.build.Build(arch, COMMIT_ID, self.include_private_scenes)
-            return (ver_build.url(), ver_build.sha256())
+            ver_build = ai2thor.build.Build(arch, COMMIT_ID, self.include_private_scenes, self.releases_dir)
+            return ver_build
         else:
-            url = None
-            sha256_build = None
             git_dir = os.path.normpath(os.path.dirname(os.path.realpath(__file__)) + "/../.git")
-            for commit_id in subprocess.check_output('git --git-dir=' + git_dir + ' log -n 10 --format=%H', shell=True).decode('ascii').strip().split("\n"):
-                commit_build = ai2thor.build.Build(arch, commit_id, self.include_private_scenes)
+            commits = subprocess.check_output('git --git-dir=' + git_dir + ' log -n 10 --format=%H', shell=True).decode('ascii').strip().split("\n")
+
+            rdir = self.releases_dir
+            found_build = None
+
+            if local_build:
+                rdir = os.path.normpath(os.path.dirname(os.path.realpath(__file__)) + "/../unity/builds")
+                commits = ['local'] + commits # we add the commits to the list to allow the ci_build to succeed
+
+            for commit_id in commits:
+                commit_build = ai2thor.build.Build(arch, commit_id, self.include_private_scenes, rdir)
 
                 try:
-                    u = commit_build.url()
-                    if os.path.isfile(self.executable_path(url=u)):
-                        # don't need sha256 since we aren't going to download
-                        url = u
-                        break
-                    elif commit_build.exists():
-                        sha256_build = commit_build.sha256()
-                        url = u
+                    if os.path.isfile(commit_build.executable_path) or (not local_build and commit_build.exists()):
+                        found_build = commit_build
                         break
                 except Exception:
                     pass
 
-            if url is None:
-                raise Exception("Couldn't find a suitable build url for platform: %s" % platform.system())
-
             # print("Got build for %s: " % (url))
 
-            return (url, sha256_build)
-
-    def build_name(self, url=None):
-        if url is None:
-            url, _ = self.build_url()
-        return os.path.splitext(os.path.basename(url))[0]
-
-    def executable_path(self, url=None):
-
-        if self.local_executable_path is not None:
-            return self.local_executable_path
-
-        target_arch = platform.system()
-
-        bn = self.build_name(url)
-        if target_arch == 'Linux':
-            return os.path.join(self.releases_dir(), bn, bn)
-        elif target_arch == 'Darwin':
-            return os.path.join(
-                self.releases_dir(),
-                bn,
-                bn + ".app",
-                "Contents/MacOS",
-                "AI2-Thor")
-        else:
-            raise Exception('unable to handle target arch %s' % target_arch)
-
-    def download_binary(self):
-
-        if self.local_executable_path:
-            return
-
-        if platform.architecture()[0] != '64bit':
-            raise Exception("Only 64bit currently supported")
-
-        url, sha256_build = self.build_url()
-        tmp_dir = os.path.join(self.base_dir(), 'tmp')
-        makedirs(self.releases_dir())
-        makedirs(tmp_dir)
-        download_lf = open(os.path.join(tmp_dir, self.build_name(url) + ".lock"), "w")
-        try:
-            fcntl.flock(download_lf, fcntl.LOCK_EX)
-
-            if not os.path.isfile(self.executable_path()):
-                zip_data = ai2thor.downloader.download(
-                    url,
-                    self.build_name(),
-                    sha256_build,
-                    self.include_private_scenes
-                    )
-
-                z = zipfile.ZipFile(io.BytesIO(zip_data))
-                # use tmpdir instead or a random number
-                extract_dir = os.path.join(tmp_dir, self.build_name())
-                logger.debug("Extracting zipfile %s" % os.path.basename(url))
-                z.extractall(extract_dir)
-                os.rename(extract_dir, os.path.join(self.releases_dir(), self.build_name()))
-                # we can lose the executable permission when unzipping a build
-                os.chmod(self.executable_path(), 0o755)
-            else:
-                logger.debug("%s exists - skipping download" % self.executable_path())
-
-        finally:
-            fcntl.flock(download_lf, fcntl.LOCK_UN)
-            download_lf.close()
-
+            return found_build
 
     def start(
             self,
@@ -914,10 +861,9 @@ class Controller(object):
 
                     self.check_x_display(env['DISPLAY'])
 
-            if not self.local_executable_path:
-                self.download_binary()
-                self.lock_release()
-                self.prune_releases()
+            self._build.download()
+            self._build.lock_sh()
+            self.prune_releases()
 
             unity_params = self.server.unity_params()
             self._start_unity_thread(env, width, height, unity_params, image_name)
@@ -934,7 +880,7 @@ class Controller(object):
         self.stop_unity()
         self.server.stop()
         self.stop_container()
-        self.unlock_release()
+        self._build.unlock()
 
     def stop_container(self):
         if self.container_id:
